# Standard library imports
from dataclasses import dataclass
from typing import TYPE_CHECKING, NamedTuple, TypedDict

# First party imports
from aeth_ext.types import StrEnum

if TYPE_CHECKING:
  # Standard library imports
  from datetime import timedelta
  from typing import NotRequired
  from zoneinfo import ZoneInfo

  # Third party imports
  from dateutil.relativedelta import relativedelta


__all__ = [
  "DEFAULT_USE_ARGS",
  "CronArgsType",
  "DayOfWeek",
  "JobID",
  "JobIDPrefix",
  "JobIDSuffix",
  "StoreNum",
  "SubJobTriggerArgs",
  "UseArgs",
]


class CronArgsType(TypedDict):
  year: NotRequired[int | str | None]
  month: NotRequired[int | str | None]
  day: NotRequired[int | str | None]
  day_of_week: NotRequired[DayOfWeek | None]
  hour: NotRequired[int | str | None]
  minute: NotRequired[int | str | None]
  second: NotRequired[int | str | None]
  timezone: NotRequired[ZoneInfo | None]


class DayOfWeek(StrEnum):
  SUNDAY = "sun"
  MONDAY = "mon"
  TUESDAY = "tue"
  WEDNESDAY = "wed"
  THURSDAY = "thu"
  FRIDAY = "fri"
  SATURDAY = "sat"


@dataclass
class UseArgs:
  year: bool = False
  month: bool = False
  day: bool = False
  day_of_week: bool = True
  hour: bool = True
  minute: bool = True
  second: bool = False


DEFAULT_USE_ARGS = UseArgs()


class SubJobTriggerArgs(NamedTuple):
  delta: timedelta | relativedelta
  use_args: UseArgs = DEFAULT_USE_ARGS


type JobID = str
type JobIDSuffix = JobID
type JobIDPrefix = JobID
type StoreNum = int


class IsPydantic: ...



# Standard library imports
from abc import abstractmethod
from contextvars import ContextVar
from datetime import datetime, timedelta
from functools import wraps
from inspect import iscoroutinefunction
from logging import getLogger
from pathlib import PurePosixPath
from typing import TYPE_CHECKING
from zoneinfo import ZoneInfo

# Third party imports
from apscheduler.triggers.cron import CronTrigger
from dateutil.relativedelta import FR, MO, SA, SU, TH, TU, WE, relativedelta
from pydantic.dataclasses import dataclass

# First party imports
from aeth_ext.errors.err_handling import FATAL_EVENT
from aeth_ext.ftp.adapter import AdaptedSFTP, FTPAdapter
from aeth_ext.types.abc import SingletonTypeABC
from aeth_ext.utils import today
from scheduled_report_aggregator.custom_types import DEFAULT_USE_ARGS, CronArgsType, DayOfWeek, IsPydantic, SubJobTriggerArgs, UseArgs
from scheduled_report_aggregator.environment_init_vars import CWD, SETTINGS
from scheduled_report_aggregator.ftp_configs import RYOSFTPClient, SASSFTPClient, SFTSFTPClient

if TYPE_CHECKING:
  # Standard library imports
  from collections.abc import Callable, Coroutine
  from datetime import timedelta
  from typing import Any, ClassVar, Unpack

  # Third party imports
  from dateutil._common import weekday

  # First party imports
  from scheduled_report_aggregator.custom_types import JobID, JobIDPrefix, JobIDSuffix
  from scheduled_report_aggregator.scheduler_config import Scheduler

logger = getLogger(__name__)


FTP_CVAR = ContextVar("FTP_CVAR")


DTUTIL_WEEKDAY_MAP: dict[DayOfWeek | None, weekday] = {
  DayOfWeek.MONDAY: MO,
  DayOfWeek.TUESDAY: TU,
  DayOfWeek.WEDNESDAY: WE,
  DayOfWeek.THURSDAY: TH,
  DayOfWeek.FRIDAY: FR,
  DayOfWeek.SATURDAY: SA,
  DayOfWeek.SUNDAY: SU,
  None: lambda x: None,  # for when day_of_week is not specified in cron args # type: ignore
}

NUM_TO_WEEKDAY_MAP: dict[int, DayOfWeek] = {
  0: DayOfWeek.MONDAY,
  1: DayOfWeek.TUESDAY,
  2: DayOfWeek.WEDNESDAY,
  3: DayOfWeek.THURSDAY,
  4: DayOfWeek.FRIDAY,
  5: DayOfWeek.SATURDAY,
  6: DayOfWeek.SUNDAY,
}


__all__ = ["HOLDING_FOLDER", "CanRescheduleJobError", "JobBase", "JobError"]
HOLDING_FOLDER = CWD / "file_holding"


@dataclass
class CronArgs(IsPydantic):
  year: int | str | None = None
  month: int | str | None = None
  day: int | str | None = None
  day_of_week: DayOfWeek | None = None
  hour: int | str | None = None
  minute: int | str | None = None
  second: int | str | None = None
  timezone: ZoneInfo | None = SETTINGS.tz

  def keys(self):
    return self.__dict__.keys()

  def __getitem__(self, key: str) -> Any:
    """Returns the value for a given field name."""
    return getattr(self, key)

  def get(self, key: str, default: Any = None) -> Any:
    return self.__dict__.get(key, default)

  def __contains__(self, key: str) -> bool:
    return key in self.__dict__


class CanRescheduleJobError(Exception):
  """Custom exception to indicate that a job should be automatically rescheduled."""

  def __init__(self, message: str, reason: str | None = None, count_error: bool = False):
    super().__init__(message)
    self.reason = reason or message
    self.count_error = count_error


class JobError(Exception):
  """Custom exception to indicate that a job has encountered an error."""

  def __init__(self, message: str, reason: str | None = None, count_error: bool = False):
    super().__init__(message)
    self.reason = reason or message
    self.count_error = count_error


class JobBase(metaclass=SingletonTypeABC):
  jobname_cvar = FTP_CVAR

  ftp_handlers: ClassVar = {
    "sft": FTPAdapter[AdaptedSFTP](SFTSFTPClient, container_cvar=FTP_CVAR),
    "sas": FTPAdapter[AdaptedSFTP](SASSFTPClient, container_cvar=FTP_CVAR),
    "ryo": FTPAdapter[AdaptedSFTP](RYOSFTPClient, container_cvar=FTP_CVAR),
  }

  errored: bool = False  # used by main to check whether this job experienced an error
  err_counter: int = 0
  err_max_threshold: int = 3  # number consecutive errors before setting errored state, triggering shutdown

  reschedule_delay_minutes: ClassVar[int] = 10  # minutes to delay when rescheduling after an error

  reports_pickup_base_folder = PurePosixPath("/upload")

  reports_pickup_folder: PurePosixPath

  jobs_register: dict[JobIDSuffix, tuple[Callable[..., Any], CronArgs | SubJobTriggerArgs]]
  extra_jobs_register: dict[JobIDSuffix, tuple[Callable[..., Any], SubJobTriggerArgs]]

  active_jobs: dict[JobID, CronArgs | SubJobTriggerArgs]

  active_args: dict[JobID, CronArgs]

  job_id: ContextVar[JobID] = ContextVar("job_id")

  base_job_id: JobIDPrefix
  scheduler: Scheduler
  jobstore: str

  def __init__(self):
    self.active_jobs = {}  # track active jobs for cleanup if needed
    self.active_args = {}  # track active jobs' trigger args for rescheduling logic
    self.extra_jobs_register = {}

    self.job_holding_folder = HOLDING_FOLDER / self.__class__.__name__.lower()
    self.job_holding_folder.mkdir(parents=True, exist_ok=True)

    self.__post_init__()  # call post init hook for any additional setup in subclasses

  @classmethod
  def init_job(
    cls,
    scheduler: Scheduler,
    job_id: JobIDPrefix,
    jobstore: str = "general_jobs",
    **kwargs: Unpack[CronArgsType],
  ) -> JobBase:
    self = cls()
    self.base_job_id = job_id
    self.scheduler = scheduler
    self.jobstore = jobstore

    self.main_cron_args = CronArgs(**kwargs)

    self.jobs_register = {
      "main_job": (self.main_job, self.main_cron_args),
    }

    return self

  def __post_init__(self): ...

  def schedule_registered_jobs(self, base_cron_args: CronArgs | None = None) -> None:
    """Hook for adding sub-jobs to the scheduler. Override in subclasses if needed."""
    now = datetime.now(tz=SETTINGS.tz)
    for job_id_suffix, (job_func, job_args) in self.jobs_register.items():
      wrapped_func, trigger, job_id = self.prep_job(job_func, job_args, job_id_suffix, base_cron_args or self.main_cron_args)
      logger.info("%s: Scheduling job '%s' to run at %s", self.__class__.__name__, job_id, trigger.get_next_fire_time(None, now))

      self.scheduler.add_job(
        wrapped_func,
        trigger=trigger,
        id=job_id,
        replace_existing=True,
        misfire_grace_time=None,  # pyright: ignore[reportArgumentType]
      )

  def prep_job(
    self,
    func: Callable[..., Any],
    trigger_args: CronArgs | SubJobTriggerArgs,
    job_id_suffix: str,
    base_cron_args: CronArgs | None = None,
  ) -> tuple[Callable[..., Any], CronTrigger, JobID]:
    """Schedules a job with the given function, trigger arguments, and job ID suffix."""
    job_id: JobID = f"{self.base_job_id}_{job_id_suffix}"

    evaled_args = (
      self.shift_cron_args(base_cron_args or self.main_cron_args, *trigger_args)  # Is sub job with delta args
      if isinstance(trigger_args, SubJobTriggerArgs)
      else trigger_args  # Is main job
    )

    trigger = CronTrigger(**evaled_args)

    # Add the job to the active jobs dict with its trigger args for tracking.
    self.active_jobs[job_id] = trigger_args
    self.active_args[job_id] = evaled_args

    # Wrap job in run_job to handle error state and rescheduling logic, then add to scheduler
    wrapped_func = self.run_job(func, job_id)

    return wrapped_func, trigger, job_id

  def run_job[**Params_T, Return_T: Any](
    self, func: Callable[Params_T, Return_T], job_id: JobID
  ) -> Callable[Params_T, Coroutine[Any, Any, Return_T | None]]:
    """Wrapper for main_job to handle error state."""

    @wraps(func)
    async def wrapper(*args: Params_T.args, **kwargs: Params_T.kwargs) -> Return_T | None:
      if self.errored:
        logger.error("%s: Job is in errored state. Skipping execution.", self.__class__.__name__)
        return

      with self.job_id.set(job_id), self.jobname_cvar.set(self.__class__.__name__):
        try:
          if iscoroutinefunction(func):
            return await func(*args, **kwargs)
          else:
            return func(*args, **kwargs)
        except CanRescheduleJobError as e:
          self.error_reschedule(count=e.count_error, reason=e.reason)

        except JobError as e:
          logger.error("%s: Job encountered a major error. Freezing this jobs execution", self.__class__.__name__, exc_info=e)
          self.errored = True
          FATAL_EVENT.set()  # trigger shutdown in main

        except Exception as e:
          logger.exception("%s: Unexpected error in main_job:", self.__class__.__name__, exc_info=e)
          self.errored = True
          FATAL_EVENT.set()  # trigger shutdown in main

    return wrapper

  @abstractmethod
  async def main_job(self) -> None:
    """Main job logic goes here. Override in subclasses."""
    raise NotImplementedError("Subclasses must implement the main_job method.")

  def cancel_self(self) -> None:
    """Cancels this job from the scheduler."""
    for job_id in self.active_jobs.copy().keys():
      self.scheduler.remove_job(job_id)
      self.active_jobs.pop(job_id, None)
      self.active_args.pop(job_id, None)

  def reset_schedule(self) -> None:
    """Resets the job's schedule to the original cron arguments."""
    self.cancel_self()
    self.schedule_registered_jobs()

  def reschedule_self(self, **kwargs: Unpack[CronArgsType]) -> None:
    """Clears this job and rebuilds it's schedule with a new base trigger."""
    self.cancel_self()
    self.main_cron_args = CronArgs(**kwargs)
    self.jobs_register["main_job"] = (self.main_job, self.main_cron_args)
    self.schedule_registered_jobs()

  def error_reschedule(self, count: bool = False, reason: str = "error in job") -> None:
    if count:
      self.err_counter += 1

      if self.err_counter >= self.err_max_threshold:
        logger.error("%s: Maximum error threshold reached. Marking job as errored and triggering shutdown.", self.__class__.__name__)
        self.errored = True
        FATAL_EVENT.set()  # trigger shutdown in main
        return

    logger.info("%s: Rescheduling due to %s", self.__class__.__name__, reason)

    delta = relativedelta(minutes=self.reschedule_delay_minutes)
    new_args = self.shift_cron_args(self.main_cron_args, delta)

    self.reschedule_self(**new_args)

  @staticmethod
  def check_if_this_week(dt: datetime) -> bool:
    now_day = today(tzinfo=SETTINGS.tz)
    start_of_week = now_day - relativedelta(weekday=SU(-1))
    end_of_week = start_of_week + relativedelta(weekday=SA(+1), hour=23, minute=59, second=59, microsecond=999999)
    return start_of_week <= dt <= end_of_week

  @staticmethod
  def extract_use_args(trigger_args: CronArgs) -> UseArgs:
    """
    Attempt to extract which cron args are being used in the provided trigger args to determine which ones to shift when rescheduling.
    If this fails, it will default to DEFAULT_USE_ARGS
    """
    try:
      return UseArgs(
        year="year" in trigger_args and trigger_args["year"] is not None,
        month="month" in trigger_args and trigger_args["month"] is not None,
        day="day" in trigger_args and trigger_args["day"] is not None,
        day_of_week="day_of_week" in trigger_args and trigger_args["day_of_week"] is not None,
        hour="hour" in trigger_args and trigger_args["hour"] is not None,
        minute="minute" in trigger_args and trigger_args["minute"] is not None,
        second="second" in trigger_args and trigger_args["second"] is not None,
      )
    except Exception:
      return DEFAULT_USE_ARGS

  def shift_cron_args(self, args: CronArgs, delta: timedelta | relativedelta, use_args: UseArgs | None = None) -> CronArgs:
    """Shifts the cron arguments by a specified timedelta."""
    if use_args is None:
      use_args = self.extract_use_args(args)

    new_cron_args = {
      "year": args.get("year") if use_args.year else None,
      "month": args.get("month") if use_args.month else None,
      "day": args.get("day") if use_args.day else None,
      # "day_of_week": args.get("day_of_week") if use_args.day_of_week else None,
      "hour": args.get("hour") if use_args.hour else None,
      "minute": args.get("minute") if use_args.minute else None,
      "second": args.get("second") if use_args.second else None,
      "tzinfo": args.get("timezone"),
    }

    reldel_args = {
      "year": args.get("year") if use_args.year else None,
      "month": args.get("month") if use_args.month else None,
      "day": args.get("day") if use_args.day else None,
      "weekday": DTUTIL_WEEKDAY_MAP[args.get("day_of_week")](-1) if use_args.day_of_week else None,
      "hour": args.get("hour") if use_args.hour else None,
      "minute": args.get("minute") if use_args.minute else None,
      "second": args.get("second") if use_args.second else None,
    }

    new_reldel = relativedelta(**reldel_args)

    # convert new_cron_args to a datetime by using the current time as a base and replacing the specified fields with the cron args values
    now = datetime.now(tz=SETTINGS.tz)
    # Filter out None values and tzinfo=True (meaning "keep existing") so datetime.replace() only receives valid args
    replace_args = {k: v for k, v in new_cron_args.items() if v is not None}
    base_dt = now.replace(**replace_args)

    shifted_dt = base_dt + new_reldel + delta

    # convert shifted_dt back to cron args by taking the relevant fields from the shifted datetime
    return CronArgs(
      year=(year if isinstance(year := args.get("year"), str) else shifted_dt.year) if use_args.year else None,
      month=(month if isinstance(month := args.get("month"), str) else shifted_dt.month) if use_args.month else None,
      day=(day if isinstance(day := args.get("day"), str) else shifted_dt.day) if use_args.day else None,
      day_of_week=NUM_TO_WEEKDAY_MAP[shifted_dt.weekday()] if use_args.day_of_week else None,
      hour=(hour if isinstance(hour := args.get("hour"), str) else shifted_dt.hour) if use_args.hour else None,
      minute=(minute if isinstance(minute := args.get("minute"), str) else shifted_dt.minute) if use_args.minute else None,
      second=(second if isinstance(second := args.get("second"), str) else shifted_dt.second) if use_args.second else None,
      timezone=args.get("timezone"),
    )



# Local folder imports
from .bal_sheet_job import BalanceSheetJob
from .base import HOLDING_FOLDER
from .timeclock_job import TimeclockJob

__all__ = ["HOLDING_FOLDER", "BalanceSheetJob", "TimeclockJob"]



# Third party imports
from pydantic import BaseModel, ConfigDict

__all__ = ["AllottedHoursModel"]


class AllottedHoursModel(BaseModel):
  model_config = ConfigDict(extra="ignore")

  store: int
  allotted_hours: int



# Standard library imports
from dataclasses import dataclass
from datetime import UTC, date, datetime
from decimal import Decimal
from io import StringIO
from logging import getLogger
from pathlib import PurePosixPath
from re import compile
from typing import TYPE_CHECKING, TypedDict, override

# Third party imports
from dateutil.relativedelta import SA, SU, relativedelta
from dateutil.rrule import DAILY, rrule
from pandas import concat, isna, read_csv, to_numeric

# First party imports
from aeth_ext.types import EmailMessageParts
from aeth_ext.utils import batch_send_emails, prepare_email_message, today
from scheduled_report_aggregator.environment_init_vars import SETTINGS

# Local folder imports
from .base import CanRescheduleJobError, JobBase

if TYPE_CHECKING:
  # Standard library imports
  from collections.abc import Callable
  from pathlib import Path
  from re import Pattern
  from typing import ClassVar, Literal

logger = getLogger(__name__)


__all__ = ["BalanceSheetJob"]


@dataclass
class FileVars:
  pickup_folder: PurePosixPath
  filename_pattern_factory: Callable[[datetime | None], Pattern[str]]
  local_holding_folder: Path

  def __post_init__(self) -> None:
    self.local_holding_folder.mkdir(parents=True, exist_ok=True)


def assemble_ryo_filename_pattern(now: datetime | None = None) -> Pattern[str]:
  now = today() if now is None else now
  dates = list(
    rrule(
      DAILY,
      dtstart=(now - relativedelta(weekday=SU(-1), hour=0, minute=0, second=0, microsecond=0)),
      until=(now + relativedelta(weekday=SA(+1), hour=23, minute=59, second=59, microsecond=999999)),
    )
  )

  years = {str(dt.year) for dt in dates}
  months = {f"{dt.month:02d}" for dt in dates}
  days = {f"{dt.day:02d}" for dt in dates}

  years_part = "|".join(years)
  months_part = "|".join(months)
  days_part = "|".join(days)

  pattern = (
    rf"^RYO_ACH_Drafts_"
    r"(?P<timestamp>"
    rf"(?P<year>{years_part})"
    rf"(?P<month>{months_part})"
    rf"(?P<day>{days_part})"
    r"(?P<hour>\d{2})"
    r"(?P<minute>\d{2})"
    r"(?P<second>\d{2})"
    r"(?P<microsecond>\d{6})"
    r")\.csv$"
  )
  return compile(pattern)


def _assemble_range_pattern(
  end_val: str | int,
  start_val: str | int = 0,
  d2_max: int = 9,
) -> str:
  """
  For a given two-digit number str (e.g. 26)
  convert it into a regex pattern (e.g. 2[0-6]|1[0-9]|0[0-9])
  Expects only 2 digits

  d2_max is inclusive
  """
  end_val = str(end_val)
  start_val = int(start_val)
  assert 0 <= start_val <= 9, "total_min must be between 0 and 9"  # noqa: PLR2004
  assert 0 <= d2_max <= 9, "d2_max must be between 0 and 9"  # noqa: PLR2004

  second_digit = None

  try:
    first_digit = int(end_val[0])
    if len(end_val) == 2:  # noqa: PLR2004
      second_digit = int(end_val[1])
  except ValueError as e:
    raise e

  if len(end_val) == 2:  # noqa: PLR2004
    patterns = []

    for d1 in range(0, first_digit + 1):
      if d1 == 0:
        patt = rf"{d1}[{start_val}-{d2_max}]"
      elif d1 == first_digit:
        patt = rf"{d1}[0-{second_digit}]"
      elif d1 > first_digit:
        raise ValueError("HOW?!")
      else:
        patt = rf"{d1}[0-9]"
      patterns.append(patt)

    return "|".join(patterns)

  elif len(end_val) == 1:
    return rf"0[{start_val}-{end_val}]"
  else:
    raise ValueError("HOW?!")


_cached_hmid = rf"({_assemble_range_pattern(end_val=23)})"

_cached_mstrt = rf"({_assemble_range_pattern(end_val=59, start_val=0)})"
_cached_mmid = rf"({_assemble_range_pattern(end_val=59)})"
_cached_mend = rf"({_assemble_range_pattern(end_val=59)})"

_cached_sstrt = rf"({_assemble_range_pattern(end_val=59, start_val=0)})"
_cached_smid = rf"({_assemble_range_pattern(end_val=59)})"
_cached_send = rf"({_assemble_range_pattern(end_val=59)})"


def assemble_sas_filename_pattern(now: datetime | None = None) -> Pattern[str]:
  now = today(tzinfo=SETTINGS.tz) if now is None else now
  start_est = now - relativedelta(weekday=SU(-1), hour=0, minute=0, second=0, microsecond=0)
  end_est = now + relativedelta(weekday=SA(+1), hour=23, minute=59, second=59, microsecond=999999)
  rrule_end_est = end_est + relativedelta(weekday=SU(+1), hour=0, minute=0, second=0, microsecond=0)
  # convert from local tz (SETTINGS.tz) to UTC
  start = start_est.astimezone(UTC)
  end = end_est.astimezone(UTC)
  rrule_end = rrule_end_est.astimezone(UTC)

  dates = list(rrule(DAILY, dtstart=start, until=rrule_end))

  days = {f"{dt.day:02d}" for dt in dates[1:-1]}

  dmid = "|".join(days)

  hstrt = rf"({_assemble_range_pattern(end_val=23, start_val=start.hour)})"
  hmid = _cached_hmid
  hend = rf"({_assemble_range_pattern(end_val=end.hour)})"

  assembled_year = r"(?P<year>{syear}|{eyear})".format(syear=start.year, eyear=end.year)  # noqa: UP032
  assembled_month = r"(?P<month>{smonth:02d}|{emonth:02d})".format(smonth=start.month, emonth=end.month)  # noqa: UP032
  assembled_day = r"(?P<day>(?P<dstrt>{sday})|(?P<dmid>{dmid})|(?P<dend>{eday}))".format(sday=start.day, dmid=dmid, eday=end.day)  # noqa: UP032
  assembled_hour = r"(?P<hour>(?(dstrt){hstrt}|(?(dmid){hmid}|{hend})))".format(hstrt=hstrt, hmid=hmid, hend=hend)  # noqa: UP032
  assembled_minute = r"(?P<minute>(?(dstrt){mstrt}|(?(dmid){mmid}|{mend})))".format(  # noqa: UP032
    mstrt=_cached_mstrt, mmid=_cached_mmid, mend=_cached_mend
  )
  assembled_second = r"(?P<second>(?(dstrt){sstrt}|(?(dmid){smid}|{send})))".format(  # noqa: UP032
    sstrt=_cached_sstrt, smid=_cached_smid, send=_cached_send
  )
  assembled_microsecond = r"(?P<microsecond>\d{1,6})"

  timestamp = r"(?P<timestamp>{year}-{month}-{day}T{hour}_{minute}_{second}\.{microsecond})".format(  # noqa: UP032
    year=assembled_year,
    month=assembled_month,
    day=assembled_day,
    hour=assembled_hour,
    minute=assembled_minute,
    second=assembled_second,
    microsecond=assembled_microsecond,
  )

  pattern = r"^Sweet_Fire_{timestamp}\.csv$".format(timestamp=timestamp)  # noqa: UP032

  return compile(pattern)


class DownloadedFiles(TypedDict):
  ryo: Path
  sas: Path


class BalanceSheetJob(JobBase):
  reschedule_delay_minutes: ClassVar[int] = 10
  email_recipients = (
    "jacob.ogden@sweetfiretobacco.com",
    "office@sweetfiretobacco.com",
  )

  def __post_init__(self) -> None:
    self.file_details = {
      "ryo": FileVars(
        pickup_folder=PurePosixPath("/Accounting"),
        filename_pattern_factory=assemble_ryo_filename_pattern,
        local_holding_folder=self.job_holding_folder / "ryo",
      ),
      "sas": FileVars(
        pickup_folder=PurePosixPath("/Outgoing/ach_detail"),
        filename_pattern_factory=assemble_sas_filename_pattern,
        local_holding_folder=self.job_holding_folder / "sas",
      ),
    }
    self.job_output_folder = self.job_holding_folder / "output"
    self.job_output_folder.mkdir(parents=True, exist_ok=True)

  @override
  async def main_job(self) -> None:
    downloaded_files = self.download_all_files()

    try:
      report_path = self.assemble_report(downloaded_files)
    except Exception as e:
      logger.exception("%s: Error assembling report:", self.__class__.__name__, exc_info=e)
      raise CanRescheduleJobError(
        "error in report assembly",
        count_error=True,
      ) from e

    try:
      self.email_report(report_path)
    except Exception as e:
      logger.exception("%s: Error emailing report:", self.__class__.__name__, exc_info=e)
      raise CanRescheduleJobError("error in emailing report", count_error=True) from e

  def download_file(self, ftp_key: str) -> Path:
    file_vars = self.file_details[ftp_key]
    with self.ftp_handlers[ftp_key].start_session() as conn:
      files = conn.listdir(file_vars.pickup_folder.as_posix())
      pattern = file_vars.filename_pattern_factory(today())

      filtered_files = filter(lambda f: pattern.match(f.filename), files)

      # check that filtered_files is not empty before calling max, otherwise it will raise a ValueError
      try:
        youngest_file = max(filtered_files, key=lambda f: f.modified_time)
      except ValueError:
        logger.warning("No matching files found in %s for FTP %s", file_vars, ftp_key)
        raise CanRescheduleJobError(
          f"Error in downloading file: missing {ftp_key} file", reason=f"missing {ftp_key} file", count_error=False
        ) from None

      remote_file = file_vars.pickup_folder / youngest_file.filename
      local_file = file_vars.local_holding_folder / youngest_file.filename
      with local_file.open("wb") as file:
        conn.download_file(remote_path=remote_file.as_posix(), callback=file.write)

    return local_file

  def download_all_files(self) -> DownloadedFiles:
    downloaded_files: dict[str, Path] = {}
    with self.jobname_cvar.set(self.__class__.__name__):
      for ftp_key in self.file_details.keys():
        try:
          local_file = self.download_file(ftp_key)
        except Exception as e:
          for file in downloaded_files.values():
            file.unlink(missing_ok=True)
          raise e

        downloaded_files[ftp_key] = local_file

    return DownloadedFiles(**downloaded_files)

  def assemble_report(self, downloaded_files: DownloadedFiles) -> Path:
    with downloaded_files["ryo"].open("r", encoding="utf-8") as ryo_file:
      ryo_first_line = ryo_file.readline()
      ryo_df = read_csv(
        ryo_file,
        header=0,
        names=["store", "ryo", "ryo_notes", "store2", "ryo2", "notes2"],
        dtype=str,
      )
    first_line_parts = ryo_first_line.strip().split(",")
    ryo_extracted_date = date.strptime(first_line_parts[1], "%m/%d/%Y")
    ryo_extracted_draft_date = date.strptime(first_line_parts[3], "%m/%d/%Y")

    ryo_df_one = ryo_df[["store", "ryo", "ryo_notes"]]
    ryo_df_two = ryo_df[["store2", "ryo2", "notes2"]]
    ryo_df_two = ryo_df_two.rename(columns={"store2": "store", "ryo2": "ryo", "notes2": "ryo_notes"})
    ryo_df = concat(
      [ryo_df_one, ryo_df_two],
      ignore_index=True,
    )

    # drop rows where store is nan
    ryo_df = ryo_df[~isna(ryo_df["store"])]

    ryo_df = ryo_df.apply(lambda col: col.str.strip())

    ryo_df.loc[:, "store"] = ryo_df["store"].str.replace(r"SFT-WHOLESALE", "SFT-WHOLESALE 999")
    ryo_cleaned_storenums = ryo_df["store"].str.extract(r"^.*?(\d+).*?$", expand=False)
    ryo_df["storenum"] = to_numeric(ryo_cleaned_storenums, errors="coerce").astype(int)

    # ryo_cleaned_amounts = ryo_df["ryo"].str.replace(r"[$,]", "", regex=True).str.replace("^-$", "0", regex=True)
    # ryo_df["ryo"] = to_numeric(ryo_cleaned_amounts, errors="coerce").astype(float)
    ryo_df["ryo_total"] = ryo_df["ryo"].str.replace(r"[$,]", "", regex=True).str.replace("^-$", "0.00", regex=True).map(Decimal)

    # ryo_df.to_csv("test_ryo.csv", header=True, index=False)

    sas_df = read_csv(
      downloaded_files["sas"],
      header=0,
      names=[
        "custnum",
        "store",
        "type",
        "invoice",
        "invoice_date",
        "draft_date",
        "amount",
        "customer_total",
      ],
      usecols=[
        "custnum",
        "store",
        "invoice",
        "invoice_date",
        "draft_date",
        "amount",
      ],
      dtype=str,
    )

    sas_df.loc[:, "storenum"] = sas_df["store"].str.extract(r"^.*?(\d+).*?$").astype(int)

    sas_df["sas_total"] = sas_df["amount"].map(Decimal)
    # sas_df.to_csv("test_sas.csv", header=True, index=False)

    # group by store and aggregate the amount column
    sas_grouped = sas_df.groupby("storenum").agg({"sas_total": "sum"}).reset_index()

    # sas_grouped.to_csv("test_sas_agged.csv", header=True, index=False)

    # join sas_total to ryo_df on storenum
    merged_df = ryo_df.merge(sas_grouped, on="storenum", how="outer")
    # merged_df.to_csv("test_merged.csv", header=True, index=True)
    merged_df = merged_df[
      [
        # "store",
        "storenum",
        "ryo_total",
        "sas_total",
        "ryo_notes",
      ]
    ]

    io_stream = StringIO(newline=None)

    io_stream.writelines(
      [
        f"RYO Date:,{ryo_extracted_date.strftime('%Y/%m/%d')}\n",
        f"RYO Draft Date:,{ryo_extracted_draft_date.strftime('%Y/%m/%d')}\n",
      ]
    )

    merged_df.to_csv(io_stream, header=True, index=False)

    now = datetime.now(tz=SETTINGS.tz)

    out_file = self.job_output_folder / f"sas_ryo_balance_sheet_{now.strftime('%Y%m%d%H%M%S%f')}.csv"

    with out_file.open("w") as report_file:
      report_file.write(io_stream.getvalue())

    return out_file

  def email_report(self, report_path: Path) -> None:
    assert isinstance(SETTINGS.alerts_email, str), "SETTINGS.alerts_email must be a string"
    msg = prepare_email_message(
      EmailMessageParts(
        subject=f"SAS/RYO Balance Sheet Report - {report_path.stem}",
        body="Attached is the latest RYO/SAS balance sheet report.",
        from_addr=("SFT Bot", None, None, SETTINGS.alerts_email),
        to_addrs=self.email_recipients,
        attachments=report_path,
      )
    )

    batch_send_emails(msg)

    logger.info("Email sent with report %s to %s", report_path.name, self.email_recipients)

  def _test_download(self, ftp_key: Literal["ryo", "sas"]) -> Path | None:
    self.download_file(ftp_key)

  def _test_assemble_report(self, downloaded_files: DownloadedFiles) -> Path:
    return self.assemble_report(downloaded_files)

  def _test_job_no_send(self) -> None:
    downloaded_files = self.download_all_files()

    try:
      report_path = self.assemble_report(downloaded_files)  # noqa: F841
    except Exception as e:
      logger.exception("%s: Error assembling report:", self.__class__.__name__, exc_info=e)
      raise CanRescheduleJobError(
        "error in report assembly",
        count_error=True,
      ) from e


if __name__ == "__main__":
  # Standard library imports
  from asyncio import run

  test_job = BalanceSheetJob()

  run(test_job.main_job())

  # result = test_job._test_download("ryo")

  # test_job._test_assemble_report(
  #   DownloadedFiles(
  #     ryo=HOLDING_FOLDER / "balancesheetjob" / "ryo" / "RYO_ACH_Drafts_20260618164600000000.csv",
  #     sas=HOLDING_FOLDER / "balancesheetjob" / "sas" / "Sweet_Fire_2026-06-17T03_31_24.476.csv",
  #   )
  # )
  # report_path = CWD / "file_holding" / "balancesheetjob" / "output" / "sas_ryo_balance_sheet_20260617093924911622.csv"

  # test_job.email_report(report_path)
